require "active_support/core_ext/module/attribute_accessors"
require "bigdecimal"
require "geld/config"
require "geld/error/currency_error"
require "geld/error/currency_not_found"
require "geld/version"

module Geld

  # Singleton Config instance that stores the convertion rates
  mattr_accessor :config

  # Main class
  class Money

    attr_accessor :amount, :currency

    # Uses bigdecimal for safer money operations
    def initialize(amount, currency)
      @amount = BigDecimal.new(amount.to_s)
      @currency = currency
    end

    # Adds a converion rate in the configuration storage
    def self.convertion_rates(currency, rates)
      Geld.config ||= Config.new
      Geld.config.add_currency(currency, rates)
    end

    # Apply the specific rate or raises error if not found
    def convert_to(new_currency)
      return dup if new_currency == self.currency

      new_amount = Geld.config &&
      Geld.config.get_rate(self.currency, new_currency) &&
      Geld.config.get_rate(self.currency, new_currency).mult(self.amount, 2)

      if new_amount.nil?
        raise Error::CurrencyNotFound.new("The desired currency is not set.")
      else
       Money.new(new_amount, new_currency)
      end
    end

    # Sums two moneys
    def +(money)
      operate(money, :add)
    end

    # Subtracts two moneys
    def -(money)
      operate(money, :sub)
    end

    # Multiplies two moneys
    def *(money)
      operate(money, :mult)
    end

    # Divides two moneys
    def /(money)
      operate(money, :div)
    end

    # Shows money formated
    def inspect
      "#{'%.2f' % amount} #{currency}"
    end

    # Tests equality
    def ==(money)
      compare(money, :==)
    end

    # Tests equality
    def !=(money)
      compare(money, :!=)
    end

    # Tests greater than
    def >(money)
      compare(money, :>)
    end

    # Tests less than
    def <(money)
      compare(money, :<)
    end

    private

    # executes an arithmetic operation after convert to the same currency
    def operate(money, operation)
      new_amount = self.amount.send(operation, money.convert_to(self.currency).amount, 2)
      Money.new(new_amount, self.currency)
    end

    # executes an comparisson after convert
    def compare(money, operator)
      self.amount.send(operator, money.convert_to(self.currency).amount)
    end

  end
end
